package hexagon.world;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.function.BiFunction;
import java.util.function.Consumer;

import org.lwjgl.input.Mouse;
import org.lwjgl.util.vector.Vector2f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.graphics.BlockRenderer;
import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.util.MathUtil;

public class World {
	
	public static final float viewDistance = 5 * (HexagonModel.y60 * 2); // Change first number only!
	public static final float reachDistance = 8f;
	public static final double angleHalfHexagon = Math.PI / (Chunk.ringSize * Chunk.chunkSize);
	public static final double radius = HexagonModel.y60 / angleHalfHexagon;
	
	private HashMap<Long, Chunk> chunks = new HashMap<>();
	private LinkedList<Block> blocksToUpdate = new LinkedList<>();
	private Block hoverBlock;
	private int blockAlteringCountdown = 0;
	
	private BiFunction<Integer, Integer, Consumer<Chunk>> tempBlockGenerator = (x, z) -> c -> {
		BlockColumn bc = null;
		for (int i = 0; i < Chunk.chunkSize; i++) {
			for (int j = 0; j < Chunk.chunkSize; j++) {
				bc = c.blockCols[i + j * Chunk.chunkSize] = new BlockColumn(c);
				if (Math.random() < 0.8) bc.setBlock(new Block(x * Chunk.chunkSize + i, 0, z * Chunk.chunkSize + j, this), 0);
				if (Math.random() < 0.9) bc.setBlock(new Block(x * Chunk.chunkSize + i, 1, z * Chunk.chunkSize + j, this), 1);
				bc.setBlock(new Block(x * Chunk.chunkSize + i, 2, z * Chunk.chunkSize + j, this), 2);
			}
		}
	};
	
	public World() {}
	
	public void tick(Camera camera) {
		updateLoadedChunks(camera);
		updateHoverBlock();
		
		int numBlocks = blocksToUpdate.size();
		while (numBlocks > 0) {
			Block b = blocksToUpdate.removeFirst();
			b.onUpdate(false);
			b.willUpdate = false;
			numBlocks--;
		}
	}
	
	private void updateLoadedChunks(Camera camera) {
		Vector2f cVec = new Vector2f(), rVec = new Vector2f(); // chunk, real
		float camX = rVec.x = camera.camX / Chunk.chunkSize;
		float camZ = rVec.y = (float) (camera.position.getCyl().rv / camera.position.getCyl().r * radius / Chunk.chunkSize);
		
		if (chunks.isEmpty()) {
			MathUtil.realCoordsToChunk(rVec, cVec);
			setNewChunk((int) cVec.x, (int) cVec.y);
		} else {
			for (Iterator<Long> it = chunks.keySet().iterator(); it.hasNext();) {
				long c = it.next();
				int cx = (int) (c >> 32), cz = (int) (c & 0xffffffff), ix, iz;
				cVec.x = cx + 0.5f;
				cVec.y = cz + 0.5f;
				MathUtil.chunkCoordsToReal(cVec, rVec);
				
				float xx, zz, distSq;
				xx = rVec.x - camX;
				zz = Math.min(zz = Math.abs(rVec.y - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
				distSq = xx * xx + zz * zz;
				
				if (distSq > viewDistance * viewDistance) {
					Chunk chunk = getChunk(cx, cz);
					chunk.onRemove();
					it.remove();
				} else {
					for (int i = 0; i < 4; i++) {
						ix = cx + (2 * i + 1) % 3 - 1;
						iz = cz + (2 * i + 1) / 3 - 1;
						iz = MathUtil.fitChunkZ(iz);
						
						if (!chunks.containsKey((long) ix << 32 | iz)) {
							cVec.x = ix + 0.5f;
							cVec.y = iz + 0.5f;
							MathUtil.chunkCoordsToReal(cVec, rVec);
							xx = rVec.x - camX;
							zz = Math.min(zz = Math.abs(rVec.y - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
							distSq = xx * xx + zz * zz;
							
							if (distSq <= viewDistance * viewDistance) {
								setNewChunk(ix, iz);
								return;
							}
						}
					}
				}
			}
		}
	}
	
	private void updateHoverBlock() {
		Main.mousePicker.updateBlockPointedAt(this);
		hoverBlock = Main.mousePicker.getHoverBlock();
		
		if (hoverBlock != null) {
			if (blockAlteringCountdown == 0) {
				if (Mouse.isButtonDown(0)) {
					removeBlock(hoverBlock.x, hoverBlock.y, hoverBlock.z);
					hoverBlock = null;
					blockAlteringCountdown = 12;
				} else if (Mouse.isButtonDown(1)) {
					int side = Main.mousePicker.getHoverSide();
					
					if (side != -1) {
						int xx = hoverBlock.x + Block.neighbourOffsets[side][0];
						int yy = hoverBlock.y + Block.neighbourOffsets[side][1];
						int zz = hoverBlock.z + Block.neighbourOffsets[side][2];
						Block block = getBlock(xx, yy, zz);
						if (block == null) {
							setBlock(xx, yy, zz, block = new Block(xx, yy, zz, this));
						}
					}
					blockAlteringCountdown = 12;
				}
			}
		}
		
		if (blockAlteringCountdown > 0) blockAlteringCountdown--;
	}
	
	public void render(Camera camera) {
		Shader.block.enable();
		
		BlockRenderer.block.begin();
		for (Chunk c : chunks.values()) {
			for (BlockColumn col : c.blockCols) {
				for (Block b : col.blocks.values()) {
					if (b != hoverBlock) {
						b.submit();
					}
				}
			}
		}
		BlockRenderer.block.end();
		BlockRenderer.block.render();
		
		if (hoverBlock != null) {
			Shader.block.setUniform1f("glow", 1.5f);
			BlockRenderer.block.begin();
			hoverBlock.submit();
			BlockRenderer.block.end();
			BlockRenderer.block.render();
			Shader.block.setUniform1f("glow", 1f);
		}
	}
	
	public Block getBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		int mx = MathUtil.superMod(x, Chunk.chunkSize), mz = z % Chunk.chunkSize;
		if (mx < 0) mx += Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cz);
		return chunk == null ? null : chunk.getBlock(mx, y, mz);
	}
	
	public void setBlock(int x, int y, int z, Block block) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		int mx = MathUtil.superMod(x, Chunk.chunkSize), mz = z % Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cz);
		if (chunk == null) chunk = setNewChunk(cx, cz);
		chunk.setBlock(mx, y, mz, block);
	}
	
	public void removeBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, Chunk.chunkSize), cz = z / Chunk.chunkSize;
		int mx = MathUtil.superMod(x, Chunk.chunkSize), mz = z % Chunk.chunkSize;
		
		Chunk chunk = getChunk(cx, cz);
		Block block;
		if (chunk != null) block = chunk.getBlock(mx, y, mz);
		chunk.removeBlock(mx, y, mz);
	}
	
	private Chunk getChunk(int x, int z) {
		return chunks.get((long) x << 32 | MathUtil.fitChunkZ(z));
	}
	
	private Chunk setNewChunk(int x, int z) {
		z = MathUtil.fitChunkZ(z);
		
		Chunk chunk = new Chunk(x, z, tempBlockGenerator.apply(x, z), this);
		Chunk prev = chunks.put((long) x << 32 | z, chunk);
		if (prev != null) prev.onRemove();
		chunk.blockCols[0].blocks.lastEntry().getValue().onUpdate(false);
		
		return chunk;
	}
	
	public void addBlockToUpdateList(Block b) {
		if (b.willUpdate) return;
		blocksToUpdate.addLast(b);
		b.willUpdate = true;
	}
	
}
