package hexagon.util;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.model.HexagonModel;
import hexagon.world.Chunk;
import hexagon.world.World;

public class MathUtil {
	
	public static final Vector3f unitX = new Vector3f(1, 0, 0);
	public static final Vector3f unitY = new Vector3f(0, 1, 0);
	public static final Vector3f unitZ = new Vector3f(0, 0, 1);
	
	private MathUtil() {}
	
	public static Matrix4f transformationMatrix(Vector2f pos, Vector2f scale, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		Matrix4f.translate(pos, matr, matr);
		Matrix4f.scale(new Vector3f(scale.x, scale.y, 1), matr, matr);
		return matr;
	}
	
	public static Matrix4f transformationMatrix(Vector3f pos, Vector3f rotation, float scale, float globalScale, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		Vector3f tempVec = new Vector3f();
		tempVec.set(globalScale, globalScale, globalScale);
		if (globalScale != 1) Matrix4f.scale(tempVec, matr, matr);
		Matrix4f.translate(pos, matr, matr);
		Matrix4f.rotate((float) (Math.PI / 2d) - rotation.x, unitX, matr, matr);
		Matrix4f.rotate(rotation.y, unitY, matr, matr);
		Matrix4f.rotate(rotation.z, unitZ, matr, matr);
		tempVec.set(scale, scale, scale);
		Matrix4f.scale(tempVec, matr, matr);
		return matr;
	}
	
	public static Matrix4f projectionMatrix(float fov, float aspect, float near, float far, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		float y_scale = (float) (1 / Math.tan(Math.PI / 180 * fov / 2) * aspect);
		float frustLen = near - far;
		
		matr.m00 = y_scale / aspect;
		matr.m11 = y_scale;
		matr.m22 = (near + far) / frustLen;
		matr.m23 = -1;
		matr.m32 = 2 * near * far / frustLen;
		matr.m33 = 0;
		
		return matr;
	}
	
	public static Matrix4f viewMatrix(Camera camera, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		Matrix4f.rotate(camera.rotation.z, unitZ, matr, matr);
		Matrix4f.rotate(camera.rotation.x, unitX, matr, matr);
		Matrix4f.rotate(camera.rotation.y, unitY, matr, matr);
		Matrix4f.rotate((float) (camera.realPosition.getCyl().rv / World.radius - Math.PI / 2), unitX, matr, matr);
		Matrix4f.translate(camera.realPosition.getEucl().negate(null), matr, matr);
		return matr;
	}
	
	public static void chunkCoordsToReal(Vector2f chunkCoords, Vector2f dest) {
		float xx = chunkCoords.x * (HexagonModel.x60 + 1);
		dest.y = fitChunkZ(chunkCoords.y + chunkCoords.x * 0.5f) * 2f * HexagonModel.y60;
		dest.x = xx;
	}
	
	public static void realCoordsToChunk(Vector2f realCoords, Vector2f dest) {
		float xx = realCoords.x / (HexagonModel.x60 + 1);
		dest.y = fitChunkZ((realCoords.y / HexagonModel.y60 - xx) / 2f);
		dest.x = xx;
	}
	
	public static float hypot(float a, float b, float c) {
		return (float) Math.sqrt(a * a + b * b + c * c);
	}
	
	public static float clamp(float a, float low, float high) {
		return a < low ? low : a > high ? high : a;
	}
	
	public static int superMod(int x, int m) {
		x %= m;
		if (x < 0) x += m;
		return x;
	}
	
	public static float superMod(float x, float m) {
		x %= m;
		if (x < 0) x += m;
		return x;
	}
	
	public static double superMod(double x, double m) {
		x %= m;
		if (x < 0) x += m;
		return x;
	}
	
	public static int fitChunkZ(int z) {
		return superMod(z, Chunk.ringSize);
	}
	
	public static float fitChunkZ(float z) {
		return superMod(z, Chunk.ringSize);
	}
	
	public static Vector3f toBlockCoords(Vector3f vec, Vector3f result, Vector3f offset) {
		if (result == null) result = new Vector3f();
		float camX = Main.main.camera.camX;
		double y = vec.y - World.radius;
		double mult = Math.pow(1 - 1 / World.radius, (Main.main.camera.camY - y) * World.xMult);
		double x = ((vec.x - camX) / mult + camX) / (HexagonModel.x60 + 1) / World.xMult;
		double z = vec.z / (2 * HexagonModel.y60) - x / 2;
		int xInt, zInt;
		double xx = x - (xInt = (int) Math.round(x));
		double zz = z - (zInt = (int) Math.round(z));
		double xp = xx + 0.5 * zz, yp = zz + 0.5 * xx, zp = yp - xp;//0.5 * (zz - xx);
		if (Math.abs(xp) > 0.5) {
			if (xp > 0) xInt++;
			else xInt--;
		} else if (Math.abs(yp) > 0.5) {
			if (yp > 0) zInt++;
			else zInt--;
		} else if (Math.abs(zp) > 0.5) {
			if (zp > 0) {
				zInt++;
				xInt--;
			} else {
				zInt--;
				xInt++;
			}
		}
		
		xx = x - xInt;
		zz = z - zInt;
		xp = xx + 0.5 * zz;
		yp = zz + 0.5 * xx;
//		zp = yp - xp;//0.5 * (zz - xx);
		
		zInt %= Chunk.totalSize;
		if (zInt < 0) zInt += Chunk.totalSize;
		zInt %= Chunk.totalSize;
		
		if (offset != null) {
			result.set(xInt, (float) Math.floor(y), zInt);
			offset.set((float) xp, (float) y - result.y - 0.5f, (float) yp);
		} else result.set((float) x, (float) y, (float) z);
		return result;
	}
	
	public static Vec3f fromBlockCoords(Vector3f vec, Vector3f offset, Vec3f result) {
		if (result == null) result = new Vec3f();
		float camX = Main.main.camera.camX;
		double mult = Math.pow(1 - 1 / World.radius, (Main.main.camera.camY - vec.y - offset.y) * World.xMult);
		result.setCyl((float) ((((HexagonModel.x60 + 1) * vec.x + offset.x) * World.xMult - camX) * mult + camX), World.radius * mult, (vec.x + 2 * vec.z + offset.z) * World.angleHalfHexagon);
		return result;
	}
	
	public static long encodeCoords(int x, int y, int z) { // bytes in long: xxxyyzzz
		return (x & 0xffffffL) << 40 | (y & 0xffffL) << 24 | z & 0xffffffL;
	}
	
	public static void floorAll(Vector3f vec) {
		vec.x = (float) Math.floor(vec.x);
		vec.y = (float) Math.floor(vec.y);
		vec.z = (float) Math.floor(vec.z);
	}
	
}
